"""
Reliqua Framework.

Copyright 2016-2024.
"""

import json
from datetime import date, datetime, time
from functools import partial

import falcon
import yaml
from falcon.media import BaseHandler
from falcon.media import JSONHandler as FalconJSONHandler
from yaml import SafeLoader


class YAMLHandler(BaseHandler):
    """Media handler class for YAML."""

    def deserialize(self, stream, _content_type, _content_length):
        """
        Load YAML from stream and return the appropriate Python object.

        :param stream:      stream to load from
        :return:            deserialized YAML as python object
        """
        try:
            return yaml.load(stream.read().decode("utf-8"), SafeLoader)
        except ValueError as exception:
            raise falcon.HTTPBadRequest(
                title="Invalid YAML",
                description=f"Could not parse YAML body: {exception}",
            ) from exception

    def serialize(self, media, _content_type):
        """
        Serialize media to YAML.

        :param media:       media to serialize
        :return:            serialized YAML
        """
        result = yaml.dump(media)

        try:
            result = result.encode("utf-8")
        except AttributeError:
            pass

        return result


class TextHandler(BaseHandler):
    """Media handler class for text."""

    def deserialize(self, stream, _content_type, _content_length):
        """
        Read text from stream.

        :param stream:      stream to load from
        :return:            text
        """
        try:
            return stream.read().decode("utf-8")
        except ValueError as exception:
            raise falcon.HTTPBadRequest(
                title="Invalid Text", description=f"Could not parse body: {exception}"
            ) from exception

    def serialize(self, media, _content_type):
        """
        Serialize media to text.

        :param media:       media to serialize
        :return:            serialized text
        """
        result = f"{media}"

        try:
            result = result.encode("utf-8")
        except AttributeError:
            pass

        return result


class JSONHandler(FalconJSONHandler):
    """
    Media handler class for JSON.

    Overloads the FalconJSONHandler to handle datetime objects.
    """

    def converter(self, obj):
        """
        Convert datetime objects to string.

        :param obj:     object to convert
        :return:        converted object
        """
        if isinstance(obj, (datetime, time, date)):
            return str(obj)
        return obj

    def __init__(self, dumps=None, loads=None):
        """Create JSONHandler instance."""
        dumps = dumps or partial(json.dumps, ensure_ascii=False, default=self.converter)
        loads = loads or json.loads
        super().__init__(dumps, loads)



"""
Reliqua framework.

Copyright 2016-2024.
"""


class Resource:
    """
    Base resource class.

    All resource should be a subclass of Resource. This is how
    the application finds and adds routes.
    """

    def __init__(self, **kwargs):
        """
        Create Resource instance.

        Create the API resource instance.

        :param dict config:      Application configuration
        :return:                 None
        """
        self.__dict__.update(kwargs)

    @property
    def name(self):
        """Return class name."""
        return self.__class__.__name__

    def get_params(self, req, keys=None, exclude=None):
        """
        Retrieve the params from the request.

        :param Request req:    The request
        :param list keys:      List of params to get
        :param list exclude:   List of params to exclude
        :return dict:          Dictionary of parameters
        """
        params = {}
        exclude = exclude or []
        for param in req.params:
            if (not keys or param in keys) and param not in exclude:
                params[param] = req.params[param]
        return params



"""
MIT License.

Copyright (c) 2017 Terrence Meiczinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


def http(code):
    """Return HTTP as string."""
    try:
        return f"{int(code)} {CODES[str(code)]}"
    except ValueError:
        return f"{MESSAGES[code.upper()]} {code}"


HTTP = http


CODES = {
    "100": "Continue",
    "101": "Switching Protocols",
    "102": "Processing",
    "200": "OK",
    "201": "Created",
    "202": "Accepted",
    "203": "Non-Authoritative Information",
    "204": "No Content",
    "205": "Reset Content",
    "206": "Partial Content",
    "207": "Multi-Status",
    "300": "Multiple Choices",
    "301": "Moved Permanently",
    "302": "Found",
    "303": "See Other",
    "304": "Not Modified",
    "305": "Use Proxy",
    "307": "Temporary Redirect",
    "308": "Permanent Redirect",
    "400": "Bad Request",
    "401": "Unauthorized",
    "402": "Payment Required",
    "403": "Forbidden",
    "404": "Not Found",
    "405": "Method Not Allowed",
    "406": "Not Acceptable",
    "407": "Proxy Authentication Required",
    "408": "Request Time-out",
    "409": "Conflict",
    "410": "Gone",
    "411": "Length Required",
    "412": "Precondition Failed",
    "413": "Payload Too Large",
    "414": "URI Too Long",
    "415": "Unsupported Media Type",
    "416": "Range Not Satisfiable",
    "417": "Expectation Failed",
    "418": "I am a teapot",
    "422": "Unprocessable Entity",
    "423": "Locked",
    "424": "Failed Dependency",
    "426": "Upgrade Required",
    "428": "Precondition Required",
    "429": "Too Many Requests",
    "431": "Request Header Fields Too Large",
    "451": "Unavailable For Legal Reasons",
    "500": "Internal Server Error",
    "501": "Not Implemented",
    "502": "Bad Gateway",
    "503": "Service Unavailable",
    "504": "Gateway Timeout",
    "505": "HTTP Version Not Supported",
    "507": "Insufficient Storage",
    "508": "Loop Detected",
    "511": "Network Authentication Required",
}

MESSAGES = {(v.upper().replace(" ", "_"), k) for k, v in CODES.items()}



"""
Reliqua Framework.

Copyright 2016-2024.
"""

import base64
import re

import falcon


def is_base64(data):
    """
    Return if string is base64.

    Check if string is base64 encoded.

    :param str data:     Base64 string
    :return bool:        True is base64
    """
    try:
        base64.b64decode(data).decode("utf-8")
    except (base64.binascii.Error, UnicodeDecodeError):
        return False

    return True


class AuthenticationContext:
    """
    Authentication context.

    A generic container for authentication details. This allows for the implementations
    for the authorization and authentication to be flexible. The minimal requirement
    is for the authentication callback to return role attribute.
    """

    role = None

    def __init__(self, **kwargs):
        """Create AuthenticationContext instance."""
        self.__dict__.update(kwargs)


class AccessControl:
    """
    Access control class.

    The abstract base class for access control. The subclass is
    expected to implement all logic.

    The `authentication_required` method returns whether a call
    requires authentication.

    The `authorized` method returns whether the client is authorized
    to execute the call. The authorized takes an `AuthenticationContext`
    which may have implentation specific details.
    """

    def authorized(self, context, _route, _method, _resource):
        """Return whether client is allowed to access resource."""
        raise NotImplementedError("authorized method not implemented")

    def authentication_required(self, _route, _method, _resource):
        """Return whether authentication is required."""
        raise NotImplementedError("authentication required method not implemented")


class AccessCallback(AccessControl):
    """
    Access callback.

    Access control is checked by calling a user defined methods. The called method
    will be supplied the endpoint, authentication context, and HTTP method. The called method
    must return `True` or `False`.
    """

    def __init__(self, authenticate_callback, authorized_callback):
        """
        Create the AccessCallback instance.

        :param callable callback:  Callback method
        :return:
        """
        self.authenticate_callback = authenticate_callback
        self.authorized_callback = authorized_callback

    def authorized(self, context, route, method, _resource):
        """
        Return whether the route or method is authorized.

        Return whether the route or method is allowed for the
        given role. The context is a generic type to allow
        for an implementation specific details.

        :param object context:  Authenticated context
        :param str route:       Route being called
        :param str method:      HTTP method invoked
        :return bool:           True if authorized
        """
        return self.authorized_callback(context, route, method)

    def authentication_required(self, route, method, _resource):
        """
        Return whether the route or method is exempt.

        Return whether the route or method requires
        authentication by calling the callback method.

        :param str route:    Route being called
        :param str method:   HTTP method invoked
        :return bool:        True if authentication is required
        """
        return self.authenticate_callback(route=route, method=method)


class AccessList(AccessControl):
    """
    Access List.

    Access control is checked by the routes and/or method specified in
    simple lists. If default mode is allow, then only routes/methods specified
    will required authentication. If default mode is deny, then only the
    routes/specified will be exempted.
    """

    def __init__(self, routes=None, methods=None, default_mode="allow"):
        """
        Create the AccessList instance.

        :param list routes:        List of routes to apply rules
        :param list methods:       List of methods to apply rules
        :param str default_mode:   Default mode (allow|deny)
        """
        self.routes = [x.lower() for x in routes]
        self.methods = [x.lower() for x in methods]
        self.default_mode = default_mode

    def authorized(self, context, _route, _method, _resource):
        """Return whether client is allowed to access resource."""
        raise NotImplementedError("authorized method not implemented")

    def authentication_required(self, route, method, _resource):
        """
        Return whether the route or method requires authentication.

        Return whether the route or method requires authentication
        by comparing the route or method against the rule lists.

        :param str route:    Route being called
        :param str method:   HTTP method invoked
        :return bool:        True if authentication is required
        """
        matched = False
        required = self.default_mode != "allowed"

        if route.lower() in self.routes or method.lower() in self.methods:
            matched = True

        # check if authentication is required
        if self.default_mode == "allow":
            # if default mode is allow, then a match means auth required
            required = not matched
        else:
            # default mode is deny, then a match means auth not required
            required = matched

        return required


class AccessMap(AccessControl):
    """
    Access Map.

    Access control is checked by the routes and/or method specified in the
    the dictionary. Only items defined will be checked and everything else will
    be denied. Therefore, this must be a complete map.

    The access map follows the form:

    route: {
        http_method: [roles]
    }

    Example:

    "/users": {
        "GET": ["admin", "user"]
    }

    You can use "*" to apply the access to all http methods.

    You can add "*" to the roles list to exempt the
    route/method from requiring authentication.
    """

    def __init__(self, access_map):
        """
        Create the AccessList instance.

        :param list access_map:    Dictionary of rules
        :return:
        """
        self.access_map = access_map

    def authorized(self, context, route, method, _resource):
        """
        Return whether client is allowed to access resource.

        :param object context:  Authentication context
        :param str route:       Route being called
        :param str method:      HTTP method being invoked
        :return bool:           True if authorized
        """
        route = self.access_map.get(route)
        roles = route.get("*") or route.get(method)
        if context.role in roles or "*" in roles:
            return True

        return False

    def authentication_required(self, route, method, _resource):
        """
        Return whether the route/method requires authentication.

        Return whether the route or method requires authentication
        by checking the resource auth dictionary.

        :param str route:    The route being called
        :param str method:   The http method invoked
        :return bool:        True if e
        """
        route = self.access_map.get(route)
        roles = route.get("*") or route.get(method)

        # if method is specified and has a wildcard role
        # then authentication is not required
        if roles and "*" in roles:
            return False

        return True


class AccessResource(AccessControl):
    """
    Access Resource Map.

    Access control is checked by the routes and/or methods specified in the resource.
    If the default mode is `allow` then routes or actions with no definition will be allowed.
    When the default mode is `deny`, then all undefine routes and methods will be denied.

    The resource map follows the form:

    __auth__: {
        http_method: [roles]
    }

    Example:

    __auth__: {
        "GET": ["admin", "user"]
    }

    You can use "*" to apply the access to all http methods.

    You can add "*" to the roles list or set roles to an empty list
    to exempt the route/method from requiring authentication.
    """

    def __init__(self, default_mode="deny", raise_on_undefined=False):
        """
        Create the AccessList instance.

        :param str default_mode:         Default mode (allow|deny)
        :param bool raise_on_undefined:  If a resource has undefined auth attributes, raise exception
        :return:
        """
        self.default_mode = default_mode
        self.raise_on_undefined = raise_on_undefined

    def authorized(self, context, _route, method, resource):
        """
        Return whether client is allowed to access resource.

        :param object context:       Authentication context
        :param str method:           HTTP method being invoked
        :param Resource resource:    Route resource
        :return bool:                True if authorized
        """
        auth = getattr(resource, "__auth__", {})
        roles = auth.get("*", []) or auth.get(method, [])

        # if no roles are defined skip authorization
        # or raise an exception
        if not roles:
            if self.raise_on_undefined:
                raise falcon.HTTPNotImplemented(f"no roles are defined for {resource.name} {method}")

            return True

        if context.role in roles or "*" in roles:
            return True

        return False

    def authentication_required(self, _route, method, resource):
        """
        Return whether the route or method requires authentication.

        Return whether the route or method for the resource requires
        authentication by comparing the route and method against the
        resources dictionary.

        :param str method:   HTTP method invoked
        :param Resource resource:    Route resource
        :return bool:       True if authentication is required
        """
        auth = getattr(resource, "__auth__", {})
        roles = auth.get(method.lower(), []) or auth.get(method.upper(), []) or auth.get("*", [])

        # If no roles are defined and default mode is allow
        # then no authentication is required.
        if not roles and self.default_mode == "allow":
            return False

        return True


class Authentication:
    """Authentication abstract base class."""

    @property
    def name(self):
        """Return authentication name."""
        return self.__class__.__name__

    def authenticate(self, _req, _resp, _resource):
        """Return whether client is authenticated."""
        raise NotImplementedError("authenticate method not implemented")


class ApiAuthentication(Authentication):
    """
    API Authentication.

    An abstract base class for API key type authentications.
    """

    location = " any"

    def __init__(self, name, description=None, validation=None):
        """
        Create an API authentication instance.

        :param str name:             Parameter name
        :param str description:      Description
        :param callable validation:  Validation callback
        :return:
        """
        self.kind = "apiKey"
        self.parameter_name = name
        self.description = description
        self.validation = validation

    @property
    def name(self):
        """Return auth name."""
        return self.__class__.__name__

    def authenticate(self, _req, _resp, resource):
        """Return whether client is authenticated."""
        raise NotImplementedError("authenticate method not implemented")

    def dict(self):
        """Return OpenAPI Schema."""
        return {
            self.name: {
                "type": self.kind,
                "name": self.parameter_name,
                "in": self.location,
                "description": self.description,
            }
        }


class CookieAuthentication(ApiAuthentication):
    """
    Cookie Authentication.

    This class provides cookie authentication. The caller must provide a
    validation callback which accepts token and returns
    user string. Any False like value will raise an error.
    """

    location = "cookie"

    def validate(self, token):
        """
        Validate the user.

        :param str token:       Client token
        :return str:            Username

        This method must be passed in during instantiation.
        """
        raise NotImplementedError("authenticate method not implemented")

    def authenticate(self, req, _resp, _resource):
        """Return whether client is authenticated."""
        api_key = req.get_cookie_values(self.parameter_name)
        if api_key:
            api_key = api_key[0]

        username = self.validation(api_key)
        if not username:
            raise falcon.HTTPUnauthorized(description="Invalid cookie authorization")

        return username


class HeaderAuthentication(ApiAuthentication):
    """
    Header Authentication.

    This class provides header authentication. The caller must provide a
    validation callback which accepts token and returns
    user string. Any False like value will raise an error.
    """

    location = "header"

    def validate(self, token):
        """
        Validate the user.

        :param str token:       Client token
        :return str:            Username

        This method must be passed in during instantiation.
        """
        raise NotImplementedError("authenticate method not implemented")

    def authenticate(self, req, _resp, _resource):
        """Return whether client is authenticated."""
        api_key = req.get_header(self.parameter_name)

        username = self.validation(api_key)
        if not username:
            raise falcon.HTTPUnauthorized(description="Invalid header authorization")

        return username


class QueryAuthentication(ApiAuthentication):
    """
    Query Authentication.

    This class provides query parameter authentication. The caller must provide a
    validation callback which accepts token and returns
    user string. Any False like value will raise an error.
    """

    location = "query"

    def validate(self, token):
        """
        Validate the user.

        :param str token:       Client token
        :return str:            Username

        This method must be passed in during instantiation.
        """
        raise NotImplementedError("authenticate method not implemented")

    def authenticate(self, req, _resp, _resource):
        """Return whether client is authenticated."""
        api_key = req.params.get(self.parameter_name)

        username = self.validation(api_key)
        if not username:
            raise falcon.HTTPUnauthorized(description="Invalid query authorization")

        return username


class BasicAuthentication(Authentication):
    """
    Basic Authentication.

    This class provides Basic authentication. The caller must provide a
    validation callback which accepts a username and password and returns
    user string. Any False like value will raise an error.
    """

    def __init__(self, validation=None):
        """
        Create BasicAuth instance.

        :param callable validation:       Callback to validate credentials
        :param AccessControl control:     Access control class
        :return:
        """
        self.kind = "http"
        self.scheme = "basic"
        self.validation = validation

    @property
    def name(self):
        """Return auth name."""
        return self.__class__.__name__

    def validate(self, username, password):
        """
        Validate the user.

        :param str username:    Client username
        :param str password:    Client password
        :return str:            Username

        This method must be passed in during instantiation.
        """
        raise NotImplementedError("authenticate method not implemented")

    def dict(self):
        """Return OpenAPI Schema."""
        return {
            self.name: {
                "type": self.kind,
                "scheme": self.scheme,
            },
        }

    def _credentials(self, req):
        header = req.get_header("Authorization")
        if not header:
            raise falcon.HTTPUnauthorized(description="Missing Authorization Header")

        m = re.search(r"Basic ([-A-Za-z0-9+/=]+)", header)
        if not m:
            raise falcon.HTTPUnauthorized(description=f"Invalid Authorization Header {header}")

        token = m.group(1)

        if not is_base64(token):
            raise falcon.HTTPUnauthorized(description="Invalid Token")

        username, _, password = base64.b64decode(token).decode("utf-8").partition(":")

        return username, password

    def authenticate(self, req, _resp, _resource):
        """
        Authenticate user.

        Authenticate with the credentials

        :return bool:     True if authenticated
        """
        username, password = self._credentials(req)

        username = self.validation(username, password)
        if not username:
            raise falcon.HTTPUnauthorized(description="Invalid authorization")

        return username


class BearerAuthentication(Authentication):
    """Bearer Token Authentication."""

    def __init__(self):
        """
        Create BearerAuth instance.

        :return:
        """
        self.kind = "http"
        self.scheme = "bearer"

    @property
    def name(self):
        """Return auth name."""
        return self.__class__.__name__

    def authenticate(self, _req, _resp, _resource):
        """Return whether client is authenticated."""
        raise NotImplementedError("authenticate method not implemented")

    def dict(self):
        """Return OpenAPI Schema."""
        return {
            self.name: {
                "type": self.kind,
                "scheme": self.scheme,
            }
        }


class MultiAuthentication(Authentication):
    """
    MultiAuth class.

    MultiAuth allows you to specify multiple auth mechanisms. They
    will then be iterated until one is successful. The access control
    specified will override any of the individual authentications.
    """

    def __init__(self, authenticators):
        """
        Create a MultiAuth instance.

        :param list[Auth] auth:    A list of Authenticators
        :return bool:               True if authenticated
        """
        self.authenticators = authenticators

    def authenticate(self, request, response, resource):
        """
        Authenticate user.

        Authenticate with the credentials

        :return bool:        True if authenticated
        """
        for auth in self.authenticators:
            try:
                return auth.authenticate(request, response, resource)
            except falcon.HTTPUnauthorized:
                pass

        raise falcon.HTTPUnauthorized(description="Invalid authorization")

    def dict(self):
        """Return OpenAPI Schema."""
        schema = {}
        for auth in self.authenticators:
            schema.update(auth.dict())

        return schema


class AuthMiddleware:
    """Auth middleware."""

    def __init__(self, authenticators, control=None):
        """
        Create a MultiAuth instance.

        :param list[Auth] auth:    A list of Authenticators
        :return bool:               True if authenticated
        """
        self.authenticators = authenticators
        self.control = control

    def dict(self):
        """Return OpenAPI Schema."""
        schema = {}
        for auth in self.authenticators:
            schema.update(auth.dict())

        return schema

    def authenticate(self, request, response, resource):
        """
        Authenticate user.

        Authenticate with the credentials

        :return bool:        True if authenticated
        """
        for auth in self.authenticators:
            try:
                return auth.authenticate(request, response, resource)
            except falcon.HTTPUnauthorized:
                pass

        raise falcon.HTTPUnauthorized(description="Invalid authorization")

    def process_resource(self, req, resp, resource, _params):
        """
        Process request after routing.

        Process the http resource. This method is called if the
        request was routed to a resource. If the user is authenticated,
        the user key will be added/set in the request context.

        :param Request req:          Request object
        :param Response resp:        Response object
        :param Resource resource:    Resource object
        :param dict params:          Additional parameters from the URI
        :return:
        """
        authorized = False

        # if resource has explicit no_auth, always skip
        # authentication.
        if getattr(resource, "no_auth", False):
            return

        # check if request requires authentication
        if not self.control.authentication_required(req.uri_template, req.method, resource):
            return

        # authenticate user
        auth = self.authenticate(req, resp, resource)

        # if an auth contecxt is returned check if authorized
        if auth:
            authorized = self.control.authorized(auth, req.uri_template, req.method, resource)
            if not authorized:
                raise falcon.HTTPUnauthorized(
                    description=f"role {auth.role} is not authorized to {req.method} {req.uri_template}"
                )

        req.context["authentication"] = auth
        req.context["authorized"] = authorized



"""
Reliqua framework.

Copyright 2016-2024.
"""

from reliqua.resources.base import Resource

from .status_codes import HTTP


def index(spec, server, sort="alpha", highlight="true"):
    """
    Return the Swagger index HTML.

    This returns the Swagger index HTML contents.

    :param str spec:       URL to the OpenAPI spec
    :param str server:     URL to the server hosting Swagger UI assets
    :param str sort:       Sort order for operations and tags
    :param str highlight:  Enable syntax highlighting
    :return str:           HTML content for Swagger UI
    """
    highlight = "true" if highlight else "false"

    return f"""<!-- HTML for static distribution bundle build -->
    <!DOCTYPE html>
    <html lang="en">
      <head>
      <meta charset="UTF-8">
        <title>Swagger UI</title>
        <link rel="stylesheet" type="text/css" href="{server}/swagger-ui.css" />
        <link rel="stylesheet" type="text/css" href="{server}/index.css" />
        <link rel="icon" type="image/png" href="{server}/favicon-32x32.png" sizes="32x32" />
        <link rel="icon" type="image/png" href="{server}/favicon-16x16.png" sizes="16x16" />
      </head>

      <body>
        <div id="swagger-ui"></div>
        <script src="{server}/swagger-ui-bundle.js" charset="UTF-8"> </script>
        <script src="{server}/swagger-ui-standalone-preset.js" charset="UTF-8"> </script>
      </body>

      <script>
          const ui = SwaggerUIBundle({{
            url: "{spec}",
            validatorUrl: null,
            dom_id: '#swagger-ui',
            deepLinking: true,
            queryConfigEnabled: true,
            presets: [
              SwaggerUIBundle.presets.apis,
              SwaggerUIStandalonePreset
            ],
            plugins: [
              SwaggerUIBundle.plugins.DownloadUrl
            ],
            layout: "StandaloneLayout",
            syntaxHighlight: {highlight},
            operationsSorter : "{sort}",
            tagsSorter: "{sort}"
          }});
      </script>
    </html>
    """


class Swagger(Resource):
    """Class to serve the static Swagger files."""

    no_auth = True

    def __init__(self, url, openapi_url, sort="alpha", highlight=True):
        """
        Create a Swagger instance.

        :param str url:            URL to Swagger instance
        :param str openapi_url:    URL to OpenAPI spec
        :param str sort:           Tag/Endpoint sort order
        :param bool highlight:     Syntax highlighting
        """
        super().__init__()
        self.url = url
        self.openapi_url = openapi_url
        self.highlight = highlight
        self.sort = sort

    def on_get(self, _req, resp):
        """
        Return the static Swagger file.

        :param Request _req:       Request object
        :param Response resp:      Response object
        """
        resp.status = HTTP(200)
        resp.content_type = "text/html"
        resp.text = index(self.openapi_url, self.url, sort=self.sort, highlight=self.highlight)



"""
MIT License.

Copyright (c) 2017 Terrence Meiczinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import glob
import importlib
import inspect
import os
import re
import sys
import uuid

import falcon
from falcon_cors import CORS

from .auth import AuthMiddleware
from .docs import Docs
from .media_handlers import JSONHandler, TextHandler, YAMLHandler
from .openapi import OpenApi
from .resources.base import Resource
from .sphinx_parser import SphinxParser
from .swagger import Swagger


class Api(falcon.App):
    """Add auto route and documentation."""

    def __init__(
        self,
        resource_path=None,
        middleware=None,
        config=None,
        resource_attributes=None,
        info=None,
        openapi=None,
        **_kwargs,
    ):
        """
        Create an API instance.

        :param str resource_path:             Path to the resource modules
        :param list middleware:               Middleware
        :param dict config:                   API configuration parameters
        :param dict resource_attributes:      Parameters added as resource attributes
        :param dict info:                     Application information
        :param dict openapi:                  OpenAPI configuration options

        :return:                               API instance
        """
        info = info or {}
        openapi = openapi or {}

        openapi_path = openapi["path"]
        self.url = openapi["ui_url"]
        self.servers = openapi["servers"]

        self.openapi = openapi
        self.info = info

        self.openapi["spec"] = f"{openapi_path}/openapi.json"
        self.openapi["static"] = f"{openapi_path}/static"
        self.resources = []
        self.auth = [x for x in middleware if isinstance(x, AuthMiddleware)]
        self.config = config or {}
        self.resource_attributes = resource_attributes or {}

        path = os.path.dirname(sys.modules[__name__].__file__)
        self.openapi["file_path"] = f"{path}/swagger"
        self.openapi["spec_url"] = f"{self.url}{self.openapi['spec']}"
        self.openapi["static_url"] = f"{self.url}{self.openapi['static']}"

        middleware = middleware or []
        cors = CORS(allow_all_origins=True, allow_all_methods=True, allow_all_headers=True)
        middleware.append(cors.middleware)

        super().__init__(middleware=middleware)

        if not resource_path:
            resource_path = path + "/resources"

        self.req_options.auto_parse_form_urlencoded = True
        self.resource_path = resource_path

        self._add_handlers()
        self._load_resources()
        self._parse_docstrings()
        self._add_routes()
        self._add_docs()

    def _add_handlers(self):
        """Add custom media handlers."""
        extra_handlers = {
            "application/yaml": YAMLHandler(),
            "text/html; charset=utf-8": TextHandler(),
            "text/plain; charset=utf-8": TextHandler(),
            "application/json": JSONHandler(),
        }

        self.req_options.media_handlers.update(extra_handlers)
        self.resp_options.media_handlers.update(extra_handlers)

    def _load_resources(self):
        """Load resource classes from the specified resource path."""
        resources = []
        path = f"{self.resource_path}/*.py"
        print(f"searching {path}")
        files = glob.glob(path)

        for file in files:
            print(f"loading {file}")
            classes = self._get_classes(file)
            resources.extend(classes)

        # Add config to resource
        self.resources = [x(app_config=self.config, **self.resource_attributes) for x in resources]

    def _is_route_method(self, name, suffix):
        """Check if a method name is a route method."""
        if not name.startswith("on_"):
            return None

        if suffix:
            return name.endswith(suffix)

        return re.search(r"^on_([a-z]+)$", name)

    def _parse_methods(self, resource, route, methods):
        """Parse methods of a resource for a given route."""
        parser = SphinxParser()
        for name in methods:
            operation_id = f"{resource.__class__.__name__}.{name}"
            action = re.search(r"on_(delete|get|patch|post|put)", name).group(1)
            method = getattr(resource, name)
            resource.__data__[route][action] = parser.parse(method, operation_id=operation_id)

    def _parse_resource(self, resource):
        """Parse a resource to extract routes and methods."""
        for route, data in resource.__routes__.items():
            resource.__data__[route] = {}
            suffix = data.get("suffix", None)
            methods = [x for x in dir(resource) if self._is_route_method(x, suffix)]
            self._parse_methods(resource, route, methods)

    def _parse_docstrings(self):
        """Parse docstrings of all resources."""
        for resource in self.resources:
            resource.__data__ = {}
            self._parse_resource(resource)

    def _get_classes(self, filename):
        """Get resource classes from a file."""
        classes = []
        module_name = str(uuid.uuid3(uuid.NAMESPACE_OID, filename))
        spec = importlib.util.spec_from_file_location(module_name, filename)
        module = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(module)
        except (ImportError, FileNotFoundError, SyntaxError, TypeError, AttributeError) as e:
            print(f"Error loading module {module_name}: {e}")
            return classes

        for _, c in inspect.getmembers(module, inspect.isclass):
            if issubclass(c, Resource) and hasattr(c, "__routes__"):
                classes.append(c)

        return classes

    def _add_routes(self):
        """Add routes for all resources."""
        for resource in self.resources:
            routes = resource.__routes__
            for route, kwargs in routes.items():
                self.add_route(route, resource, **kwargs)

    def _add_docs(self):
        """Add Swagger and OpenAPI documentation routes."""
        swagger = Swagger(
            self.openapi["static_url"],
            self.openapi["spec_url"],
            sort=self.openapi["sort"],
            highlight=self.openapi["highlight"],
        )
        openapi = OpenApi(**self.info, auth=self.auth, servers=self.servers)
        openapi.process_resources(self.resources)
        schema = openapi.schema()
        print(f"adding static route {self.openapi['docs']} {self.openapi['file_path']}")
        self.add_static_route(self.openapi["static"], self.openapi["file_path"])
        self.add_route(self.openapi["docs"], swagger)
        print(f"adding openapi file {self.openapi['spec']}")
        self.add_route(self.openapi["spec"], Docs(schema))



"""
MIT License.

Copyright (c) 2017 Terrence Meiczinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from falcon.http_error import HTTPError


class HTTPBadRequest(HTTPError):
    """Bad Request."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "400 Bad Request"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPUnauthorized(HTTPError):
    """Unauthorized."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "401 Unauthorized"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPPaymentRequired(HTTPError):
    """Payment Required."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "402 Payment Required"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPForbidden(HTTPError):
    """Forbidden."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "403 Forbidden"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPNotFound(HTTPError):
    """Not Found."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "404 Not Found"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPMethodNotAllowed(HTTPError):
    """Method Not Allowed."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "405 Method Not Allowed"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPNotAcceptable(HTTPError):
    """Not Acceptable."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "406 Not Acceptable"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPProxyAuthenticationRequired(HTTPError):
    """Proxy Authentication Required."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "407 Proxy Authentication Required"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPRequestTimeout(HTTPError):
    """Request Timeout."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "408 Request Timeout"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPConflict(HTTPError):
    """Conflict."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "409 Conflict"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPGone(HTTPError):
    """Gone."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "410 Gone"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPLengthRequired(HTTPError):
    """Length Required."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "411 Length Required"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPPreconditionFailed(HTTPError):
    """Precondition Failed."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "412 Precondition Failed"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPPayloadTooLarge(HTTPError):
    """Payload Too Large."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "413 Payload Too Large"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPUnsupportedMediaType(HTTPError):
    """Unsupported Media Type."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "415 Unsupported Media Type"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPRangeNotSatisfiable(HTTPError):
    """Range Not Satisfiable."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "416 Range Not Satisfiable"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPExpectationFailed(HTTPError):
    """Expectation Failed."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "417 Expectation Failed"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPUnprocessableEntity(HTTPError):
    """Unprocessable Entity."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "422 Unprocessable Entity"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPLocked(HTTPError):
    """Locked."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "423 Locked"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPFailedDependency(HTTPError):
    """Failed Dependency."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "424 Failed Dependency"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPUpgradeRequired(HTTPError):
    """Upgrade Required."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "426 Upgrade Required"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPPreconditionRequired(HTTPError):
    """Precondition Required."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "428 Precondition Required"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPTooManyRequests(HTTPError):
    """Too Many Requests."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "429 Too Many Requests"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPRequestHeaderFieldsTooLarge(HTTPError):
    """Request Header Fields Too Large."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "431 Request Header Fields Too Large"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPUnavailableForLegalReasons(HTTPError):
    """Unavailable For Legal Reasons."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "451 Unavailable For Legal Reasons"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPInternalServerError(HTTPError):
    """Internal Server Error."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "500 Internal Server Error"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPNotImplemented(HTTPError):
    """Not Implemented."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "501 Not Implemented"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPBadGateway(HTTPError):
    """Bad Gateway."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "502 Bad Gateway"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPServiceUnavailable(HTTPError):
    """Service Unavailable."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "503 Service Unavailable"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPGatewayTimeout(HTTPError):
    """Gateway Timeout."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "504 Gateway Timeout"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPVersionNotSupported(HTTPError):
    """HTTP Version Not Supported."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "505 HTTP Version Not Supported"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPInsufficientStorage(HTTPError):
    """Insufficient Storage."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "507 Insufficient Storage"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPLoopDetected(HTTPError):
    """Loop Detected."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "508 Loop Detected"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )


class HTTPNetworkAuthenticationRequired(HTTPError):
    """Network Authentication Required."""

    def __init__(self, title=None, description=None, headers=None):
        """Initialize."""
        status = "511 Network Authentication Required"
        title = title or status

        super().__init__(
            status=status,
            title=title,
            description=description,
            headers=headers,
        )



"""
Reliqua Framework.

Copyright 2016-2024.
"""

from reliqua.resources.base import Resource


class Docs(Resource):
    """
    Documents endpoint class.

    This class will process resource method doc strings to generate a schema
    that will be used for the API, parameter checking and validation.

    Example:

    Description of method.

    :param str username:  [in_path, required] User ID
    :param str email:     [in_query] User Email
    :param str phone:     [in_query, enum=phones] Phone Numbers

    :response 200:        user was retrieved
    :response 400:        invalid query parameter

    :return json:

    -- Parameters --
    Parameters follow standard Python type. Modifier options are
    placed within []'s and are followed by the parameter description.

    Modifiers are only needed to override default values:

    in_query|in_path|in_body      Where parameter will be read from. [default: in_query]
    required                      Whether parameter is required. [default: False (except POST)]
    enum                          The parameter values are limited by a list.
                                  The enum values will be retrieved from the resource as follows:
                                      if in the form enum=<name>, then name
                                      plural version of parameter name
                                      singular version of the parameter name
    -- Responses --
    By default all standard HTTP messages will be available as defined by status codes. They only
    need to be listed here to change the message or to explicitly show the message in the API
    documentation.

    -- Return --
    The return type of the method.
    """

    no_auth = True

    def __init__(self, schema):
        """
        Create Docs instance.

        :param dict schema:    Documents JSON schema
        :return:               None
        """
        super().__init__()
        self.schema = schema

    def on_get(self, _req, resp):
        """
        Return the JSON document schema.

        :param Request _req:       Request object
        :param Response resp:      Response object
        :return:                   None
        """
        resp.set_header("Access-Control-Allow-Origin", "*")
        resp.media = self.schema



"""
Reliqua Framework.

Copyright 2016-2024.
"""

import binascii
import os
from base64 import b64decode, b64encode

from peewee import (
    Model,
    MySQLDatabase,
    PostgresqlDatabase,
    Proxy,
    SqliteDatabase,
)

db = Proxy()


def is_b64(string):
    """
    Check if string is base64 encoded.

    :param str s:   String to check
    :return:        True if string is base64 encoded
    """
    try:
        return b64encode(b64decode(string)) == string
    except (ValueError, TypeError, binascii.Error):
        return False


class BaseModel(Model):
    """Base model class for applications using a PeeWee DB ORM."""

    class Meta:
        """Base metaclass."""

        database = db


class DatabaseConnection:
    """
    This middleware will handle the session connect and disconnect to the database.

    This is primarily needed for MySQL databases where the connections
    timeout, but the session does not.
    """

    def __init__(self, database):
        """Create DatabaseConnection instance."""
        self.db = database

    def process_request(self, _req, _resp):
        """
        Process the request.

        :param req:                 Request object
        :param resp:                Response object
        """
        self.db.connect()

    def process_response(self, _req, _resp, _resource, _req_succeeded):
        """
        Process the response.

        :param req:                 Request object
        :param resp:                Response object
        :param resource:            Resource object
        :param bool req_succeeded:  ``True`` if request succeeded else ``False``
        """
        if not self.db.is_closed():
            self.db.close()


def mysql_connect(host=None, database=None, user=None, password=None, port=3306):
    """
    Connect to MySQL database.

    :param str database:    Database Name
    :param str user:        User
    :param str password:    Password
    :param str host:        Host
    :param str port:        Port

    :return:                MySQLDatabase
    """
    if is_b64(password):
        password = b64decode(password)
    mysql = MySQLDatabase(database, user=user, password=password, host=host, port=port)
    db.initialize(mysql)

    return db


def psql_connect(host=None, database=None, user=None, password=None, port=5432):
    """
    Connect to PostgreSQL database.

    :param str database:    Database Name
    :param str user:        User
    :param str password:    Password
    :param str host:        Host
    :param str port:        Port

    :return:                PostgresqlDatabase
    """
    if is_b64(password):
        password = b64decode(password)
    mysql = PostgresqlDatabase(database, user=user, password=password, host=host, port=port)
    db.initialize(mysql)

    return db


def sqlite_connect(path):
    """
    Connect to SQLite database.

    :param str path:       Path to database file

    :return:               SqliteDatabase
    """
    path = os.path.expanduser(path)
    database = SqliteDatabase(path)
    db.initialize(database)

    return db



"""
Reliqua Framework.

Copyright 2016-2024.
"""

import json
import re

import falcon

TRANSFORMS = {
    "string": str,
    "str": str,
    "number": float,
    "float": float,
    "integer": int,
    "int": int,
    "boolean": bool,
    "bool": bool,
    "object": json.loads,
    "list": list,
    "array": list,
    "dict": dict,
}


def python_type(s):
    """Return Python type from string."""
    return __builtins__.get(s)


class Parameter:
    """
    Parameter class.

    This class represents a python typed parameter.
    """

    def __init__(self, **kwargs):
        """
        Create a parameter instance.

        Creates a parameter object and set attributes.
        """
        self.default = None
        self.required = False
        self.__dict__.update(kwargs)


class Converter:
    """
    Converter class.

    This class contains various str to Python type conversions.
    """

    @staticmethod
    def as_str(req, name, default=None, required=False, **_kwargs):
        """
        Get the parameter as a string.

        :param req:             Request object
        :param str name:        Name of the parameter
        :param str default:     Default value of the parameter
        :param bool required:   ``True`` if the parameter is required else ``False``
        :return str:            Converted parameter value
        """
        return req.get_param(name, default=default, required=required)

    @staticmethod
    def as_float(req, name, default=None, required=False, **_kwargs):
        """
        Get the parameter as a float.

        :param req:             Request object
        :param str name:        Name of the parameter
        :param float default:   Default value of the parameter
        :param bool required:   ``True`` if the parameter is required else ``False``
        :return float:          Converted parameter value
        """
        default = float(default) if default else None
        return req.get_param_as_float(name, default=default, required=required)

    @staticmethod
    def as_int(req, name, default=None, required=False, **_kwargs):
        """
        Get the parameter as an integer.

        :param req:             Request object
        :param str name:        Name of the parameter
        :param int default:     Default value of the parameter
        :param bool required:   ``True`` if the parameter is required else ``False``
        :return int:            Converted parameter value
        """
        default = int(default) if default else None
        return req.get_param_as_int(name, default=default, required=required)

    @staticmethod
    def as_bool(req, name, default=None, required=False, **_kwargs):
        """
        Get the parameter as a boolean.

        :param req:             Request object
        :param str name:        Name of the parameter
        :param bool default:    Default value of the parameter
        :param bool required:   ``True`` if the parameter is required else ``False``
        :return bool:           Converted parameter value
        """
        return req.get_param_as_bool(name, default=default, required=required)

    @staticmethod
    def as_object(req, name, default=None, required=False, **_kwargs):
        """
        Get the parameter as an object.

        :param req:             Request object
        :param str name:        Name of the parameter
        :param dict default:    Default value of the parameter
        :param bool required:   ``True`` if the parameter is required else ``False``
        :return:                Converted parameter value
        """
        # parameter data is already converted
        if isinstance(req.params[name], dict):
            return req.params[name]

        return req.get_param_as_json(name, default=default, required=required)

    @staticmethod
    def as_list(req, name, default=None, required=False, transform=None, **_kwargs):
        """
        Get the parameter as a list.

        :param req:             Request object
        :param str name:        Name of the parameter
        :param list default:    Default value of the parameter
        :param bool required:   ``True`` if the parameter is required else ``False``
        :param transform:       Function to transform the list items
        :return:                Converted parameter value
        """
        # if param is a string, convert to list and strip whitespace
        # handles cases where commas were encoded and bypassed Falcon's built-in conversion
        # also removes empty strings from list
        if isinstance(req.params[name], str) and req.params[name]:
            value = req.params[name].strip("'\"").split(",")
            req.params[name] = [x.strip() for x in value]

        items = req.get_param_as_list(name, default=default, required=required, transform=transform)

        return [item for item in items if item != ""]

    @staticmethod
    def as_array(req, name, default=None, required=False, transform=None, **_kwargs):
        """
        Get the parameter as a list.

        :param req:             Request object
        :param str name:        Name of the parameter
        :param list default:    Default value of the parameter
        :param bool required:   ``True`` if the parameter is required else ``False``
        :param transform:       Function to transform the list items
        :return:                Converted parameter value
        """
        Converter.as_list(req, name, default=default, required=required, transform=transform)

    @staticmethod
    def convert(req, parameter, transform=None):
        """
        Convert parameter types to Python types.

        :param Request req:          Request object
        :param Parameter parameter:  Parameter object
        :param str transform:        Transform method name
        :return:                     None
        """
        # pre-check if already type converted and skip if needed
        try:
            value = req.get_param(parameter.name, required=parameter.required)
        except IndexError:
            value = []

        if isinstance(value, python_type(parameter.datatype)):
            return value

        converter = getattr(Converter, f"as_{parameter.datatype}", Converter.as_str)
        transform = TRANSFORMS.get(transform, str)
        default = transform(parameter.default) if transform and parameter.default else None

        return converter(
            req,
            parameter.name,
            default=default,
            required=parameter.required,
            transform=transform,
        )


class ProcessParams:
    """This middleware will process parameters and convert them to python types."""

    def _check_required(self, request, parameter):
        """
        Check if specified parameter is required.

        If the specified parameter is required and not present,
        raise the appropriate HTTP error.

        :param Request request:         Request object:
        :param Parameter parameter:     Parameter object
        :return:                        None
        :raises falcon.HTTPBadRequest:  Falcon bad request exception
        """
        value = parameter.default or request.params.get(parameter.name)
        if parameter.required and not value:
            raise falcon.HTTPBadRequest(
                title="Bad Request",
                description=f"Missing parameter '{parameter.name}'",
            )

    def _convert(self, request, parameter):
        pass

    def _parse_operators(self, request):
        operators = {}
        name = ""

        for param in list(request.params.keys()):
            name, _, operator = param.partition("__")
            if operator:
                operators[name] = operator
                request.params[name] = request.params.pop(param)

        return operators

    @staticmethod
    def get_resource_parameters(request, resource):
        """
        Return resource parameters.

        Return the resource parameter data dictionary.

        :param Request request:            Request object
        :param Response response:          Response object
        :return list[dict]                 Return list of parameter dictionaries
        """
        try:
            endpoint = request.uri_template
            method = request.method.lower()
            params = resource.__data__[endpoint][method]["parameters"]
        except (TypeError, AttributeError, KeyError):
            params = []

        return [Parameter(**x) for x in params]

    def process(self, request, parameters):
        """
        Process parameters.

        Process, validate, and type convert to python types.

        :param Request request:             Request object
        :param list[Parameter] parameters:  List of parameters from the schema
        :return:                            None
        """
        operators = self._parse_operators(request)

        # include operator dict in params if operators are found
        if operators:
            request.params["operators"] = operators

        # use the docs schema to validate
        for parameter in parameters:
            transform = None
            present = parameter.name in request.params

            # check for required parameters
            self._check_required(request, parameter)

            # if parameter is not required, not specified, and has no default, move on
            if not present and not parameter.default:
                continue

            # for operators in and between, datatype must be a list
            if operators.get(parameter.name) in ["in", "between"]:
                if "list" not in parameter.datatype:
                    parameter.datatype = "list"

            if m := re.search(r"list\[(\w+)]", parameter.datatype):
                parameter.datatype = "list"
                transform = m.group(1)

            request.params[parameter.name] = Converter.convert(
                request,
                parameter,
                transform=transform,
            )

    def process_form(self, form):
        """
        Process form data.

        Process the fields from the form data.

        :param MultipartForm form:    The multi-part form
        """
        data = {}

        for part in form:
            if part.content_type == "text/plain":
                data[part.name] = part.text
            if part.content_type == "application/json":
                data[part.name] = part.get_media()

        return data

    def process_resource(self, request, _response, resource, params):
        """
        Process resource.

        Process  the parameters for the request resource.

        :param Request request:     Request object
        :param Response _response:  Response object
        :param Resource resource:   Resource object
        :param list params:         Resource parameters
        :return:                    None
        """
        schema = self.get_resource_parameters(request, resource)

        # combine parameters from query, path, and body
        for media_params in [params, request.get_media(default_when_empty=None)]:
            if isinstance(media_params, dict):
                request.params.update(media_params)
            elif isinstance(media_params, falcon.media.multipart.MultipartForm):
                media_params = self.process_form(media_params)
                request.params.update(media_params)

        # if resource has no schema, skip further processing
        if not schema:
            return

        # process, validate, convert, etc
        self.process(request, schema)



"""
Reliqua Framework.

Copyright 2016-2024.
"""

import configparser

from gunicorn.app.base import BaseApplication

from .api import Api
from .middleware import ProcessParams

GUNICORN_DEFAULTS = {
    "bind": "127.0.0.1:8000",  # Address and port to listen for requests [host:port]
    "workers": 1,  # Number of worker processes
    "loglevel": "critical",  # Log level (debug, error, info, critical)
    "accesslog": "-",  # Access log path ("-" for stream, None to disable)
    "errorlog": "-",  # Error log path ("-" for stream, None to disable)
    "worker_class": "gthread",  # Type of worker processes
    "timeout": 30,  # Inactive worker timeout (worker killed)
    "keepalive": 2,  # Keep alive period for a request
}

OPENAPI_DEFAULTS = {
    "ui_url": None,  # Swagger UI index URL (if different from bind, for example, behind a proxy))
    "highlight": True,  # Enable OpenAPI syntax highlighting
    "sort": "alpha",  # OpenAPI endpoint/tag sort order
    "path": "/openapi",  # OpenAPI endpoint path (JSON and static files)
    "docs": "/docs",  # Documentation endpoint
    "servers": [],  # List of target API servers (default is bind address)
}

INFO_DEFAULTS = {
    "title": "Application Title",  # Application title
    "version": "1.0.0",  # Application version
    "description": "",  # Application description
    "license": "",  # License
    "license_url": "",  # License URL
    "contact_name": "",  # Contact name
}


def update_dict(a, b):
    """
    Update dictionary a with values from dictionary b.

    Update dictionary a with values from dictionary b. Remove
    keys from a that are not in b.

    :param dict a:  Dictionary to update
    :param dict b:  Dictionary with new values
    :return dict:   Updated dictionary
    """
    # Create a new dictionary with keys from a that are also in b
    updated_dict = {key: a[key] for key in a if key in b}

    # Add keys from b that are not in a
    for key in b:
        if key not in updated_dict:
            updated_dict[key] = b[key]

    return updated_dict


def load_config(config_file):
    """
    Load configuration file.

    :param str config_file:    Configuration file
    :return dict:              Options dictionary
    """
    params = {}

    try:
        section = "config"
        config = configparser.ConfigParser()
        config.read(config_file)

        for option in config.options(section):
            params[option] = config.get(section, option)
    except (TypeError, configparser.Error) as e:
        # Log the error or handle it appropriately
        print(f"Error loading config file: {e}")

    return params


class Application(BaseApplication):
    """Create a standalone API application."""

    def __init__(
        self,
        resource_path=None,
        middleware=None,
        config=None,
        resource_attributes=None,
        info=None,
        openapi=None,
        gunicorn=None,
        **_kwargs,
    ):
        """
        Create Application instance.

        :param str resource_path:         Path to the API resource modules
        :param list middleware:           Middleware
        :param dict config:               Application resource parameters (passed to resources as config attribute)
        :param dict resource_attributes:  Parameters added as resource attributes
        :param str info:                  Application information
        :param str openapi:               OpenAPI configuration options
        :param dict gunicorn:             Gunicorn configuration options

        :return:                          Application instance
        """
        middleware = middleware or []
        gunicorn = gunicorn or {}
        openapi = openapi or {}
        info = info or {}

        resource_path = resource_path or ""
        resource_attributes = resource_attributes or {}

        self.gunicorn_options = update_dict(gunicorn, GUNICORN_DEFAULTS)
        openapi = update_dict(openapi, OPENAPI_DEFAULTS)
        info = update_dict(info, INFO_DEFAULTS)

        middleware.append(ProcessParams())

        # Trim slashes from proxy URL if specified; otherwise set default proxy URL
        bind = self.gunicorn_options["bind"]
        openapi["ui_url"] = openapi["ui_url"].rstrip("/") if openapi["ui_url"] else f"http://{bind}"

        # Add default api server if none specified
        if len(openapi["servers"]) == 0:
            openapi["servers"] = [{"url": openapi["ui_url"], "description": "Default server"}]

        self.application = Api(
            resource_path=resource_path,
            middleware=middleware,
            config=config,
            resource_attributes=resource_attributes,
            openapi=openapi,
            info=info,
        )

        super().__init__()

    def init(self, _parser, _opts, _args):
        """
        Init method.

        This is not invoked, since the load_config method
        is also overridden.
        """
        return

    def load_config(self):
        """
        Load configuration.

        Default config loader. This load settings from a config file
        with a 'config' section. This method should be overloaded
        if a custom loader is required.
        """
        for key, value in self.gunicorn_options.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        """
        Load the application.

        Base class.
        """
        return self.application



"""
Reliqua Framework.

Copyright 2016-2024.
"""

import inspect
import re

PARAM_REGEX = re.compile(
    r":param\s+(?P<datatype>[\[\]|\w]+)\s+(?P<name>[\S+]+):\s+(\[(?P<options>.*)\])?\s+(?P<description>.*)"
)
PARAM_ITER_REGEX = re.compile(
    r"^(:param.*?:.*?)(?:(?=:param)|(?=:return)|(?=:response)|(?=:accepts))",
    re.MULTILINE | re.DOTALL,
)
RESPONSE_ITER_REGEX = re.compile(r":response\s+(?P<code>\d+)\s*(?P<schema>\w+)?:\s+(?P<description>.*)")
RETURN_REGEX = re.compile(r":return[s]*\s*(\[(.*?)\]|\w+)")
ACCEPT_REGEX = re.compile(r":accepts\s*(\[(.*?)\]|\w+)")
KEYVALUE_REGEX = re.compile(r"(?P<key>\w+)=(?P<value>\S+)")
OPERATION_REGEX = re.compile(r"on_(delete|get|patch|post|put)")
SUFFIX_REGEX = re.compile(r"on_(?:delete|get|patch|post|put)_([a-zA-Z0-9_]+)")


class SphinxParser:
    """
    Sphinx docstring parser.

    This class contains methods to parse resource's method docstring and create
    a structures schema.
    """

    def __init__(self):
        """
        Create SphinxParser instance.

        Create a SphinxParser instance to parse resource doc strings.
        """
        self.doc = ""

    @staticmethod
    def parse_options(string):
        """
        Parse options docstring.

        Parse option information from the docstring.

        :param str string:    Docstring
        :return dict:         Options dict
        """
        options = {
            "location": "query",
            "enum": [],
            "required": False,
            "default": None,
            "min": None,
            "max": None,
        }
        for match in KEYVALUE_REGEX.finditer(string):
            items = match.groupdict()
            options[items["key"]] = items["value"]

        # rename in to location
        options["location"] = options.pop("in", "query")
        options["required"] = "required" in string and "optional" not in string

        return options

    def parse_parameter(self, string):
        """
        Parse parameter docstring.

        Parse the parameter information from the docstring.

        :param str string:     Docstring
        :return dict:          Parameter dict
        """
        match = PARAM_REGEX.search(string)

        if not match:
            return None

        parameter = match.groupdict()
        if options := parameter.pop("options"):
            parameter.update(self.parse_options(options))

        # set required default based on type
        if not parameter.get("required"):
            parameter["required"] = False

        parameter["description"] = parameter["description"].strip()

        return parameter

    @property
    def summary(self):
        """Return summary."""
        if match := re.search(r"(.*?)\n", self.doc, re.MULTILINE | re.DOTALL):
            return match.group(1).replace("\n", " ").strip()

        return ""

    @property
    def description(self):
        """Return description."""
        if match := re.search(r"\n\n(.*?):", self.doc, re.MULTILINE | re.DOTALL):
            return match.group(1).replace("\n", " ").strip()

        return ""

    @property
    def parameters(self):
        """Return parameters."""
        _parameters = []

        for item in PARAM_ITER_REGEX.finditer(self.doc):
            string = " ".join([x.strip() for x in item.group(1).splitlines()])
            parameter = self.parse_parameter(string)
            if parameter:
                _parameters.append(parameter)

        return _parameters

    @property
    def accepts(self):
        """Return accepts type."""
        m = ACCEPT_REGEX.search(self.doc)

        if not m:
            return []

        parsed = m.group(1).strip("[]") or "json"
        return re.split(r",\s*|\s+", parsed)

    @property
    def responses(self):
        """Return responses."""
        _responses = []

        for item in RESPONSE_ITER_REGEX.finditer(self.doc):
            response = item.groupdict()
            response["schema"] = response.get("schema", {})
            _responses.append(response)

        return _responses

    @property
    def content(self):
        """Return content."""
        m = RETURN_REGEX.search(self.doc)

        if not m:
            return ["json"]

        parsed = m.group(1).strip("[]") or "json"
        return re.split(r",\s*|\s+", parsed)

    @staticmethod
    def generate_operation_id(method):
        """
        Generate operation id.

        Generate an operation id from the method name.

        :param method method:    Python method
        :return str:             Operation id
        """
        return method.__qualname__

    @staticmethod
    def parse_operation(method):
        """
        Parse operation.

        Parse the operation information from the docstring.

        :param method method:    Python method
        :return dict:            Dict of parameter options
        """
        m = re.search(OPERATION_REGEX, method.__qualname__)
        if m:
            return m.group(1)

        return None

    @staticmethod
    def _parse_suffix(method):
        m = re.search(SUFFIX_REGEX, method.__qualname__)
        if m:
            return m.group(1)

        return None

    def parse(self, method, operation=None, operation_id=None):
        """
        Parse the methods docstring.

        :param method method:     Method with docstring
        :param str operation:     Operation type (default is to parse from method name)
        :param str operation_id:  Operation id (default is to parse from method name)
        :return dict:             Return operation dictionary
        """
        self.doc = inspect.cleandoc(method.__doc__)
        operation = operation or self.parse_operation(method)
        operation_id = operation_id or self.generate_operation_id(method)
        suffix = self._parse_suffix(method)

        return {
            "operation": operation,
            "operation_id": operation_id,
            "suffix": suffix,
            "summary": self.summary,
            "description": self.description,
            "parameters": self.parameters,
            "responses": self.responses,
            "return_types": self.content,
            "accepts": self.accepts,
        }



"""
Reliqua Framework.

Copyright 2016-2024.
"""

import re


def camelcase(string):
    """
    Convert underscore names to camel case.

    :param str string:        String to convert
    :return str:              Camel case string
    """
    return "".join(word.title() if i else word for i, word in enumerate(string.split("_")))


CONTENT_MAP = {
    "binary": "application/octet-stream",
    "json": "application/json",
    "yaml": "application/yaml",
    "xml": "application/xml",
    "gzip": "application/gzip",
    "html": "text/html; charset=utf-8",
    "text": "text/plain; charset=utf-8",
    "jpeg": "image/jpeg",
    "png": "image/png",
    "gif": "image/gif",
    "form": "multipart/form-data",
    "*/*": "*/*",
}


BINARY_TYPES = ["binary", "gzip", "jpeg", "gif"]


TYPE_MAP = {
    "str": "string",
    "int": "integer",
    "float": "number",
    "bool": "boolean",
    "json": "object",
    "object": "object",
    "list": "array",
    "dict": "object",
}


DEFAULT_RESPONSE = {
    "type": "object",
    "properties": {
        "message": {
            "type": "string",
            "examples": [
                "example message",
            ],
        },
    },
}


BINARY_RESPONSE = {"type": "string", "format": "binary"}


verbs = ["get", "patch", "put", "post", "delete"]


class Parameter:
    """
    OpenAPI Parameter class.

    This class represents an OpenAPI parameter.
    """

    def __init__(
        self,
        name=None,
        datatype=None,
        location=None,
        required=None,
        default=None,
        enum=None,
        description=None,
        explode=None,
        format=None,
        min=None,
        max=None,
        examples=None,
    ):
        """
        Create Parameter instance.

        Creates a parameter object.

        :param str name:        Name
        :param str datatype:    Parameter type
        :param str location:    Where parameter is located
        :param bool required:   Required parameter
        :param bool default:    Default value
        :param str enum:        Variable name of the enums
        :param str description: Parameter description
        :param bool explode:    Whether arrays and objects should generate separate parameters
        :param str format:      Parameter format type
        :param int min:         Min value
        :param int max:         Max value
        :param str examples:    Example value
        """
        self.name = name
        self.location = location or "query"
        self.required = required
        self.default = default
        self.enum = enum
        self.description = description
        self.explode = explode
        self.format = format
        self.min = min
        self.max = max
        self.examples = examples
        self._datatype = datatype

        # Path parameters are always required
        if self.location == "path":
            self.required = True

    @property
    def datatype(self):
        """Return the parameter datatype."""
        value = self._datatype
        a, _, b = value.partition("|")

        # Handle defined list type ex: list[str]
        if re.search(r"list\[(\w+)]", a):
            a = "list"

        return f"{TYPE_MAP[a]}|{TYPE_MAP[b]}" if b else f"{TYPE_MAP[a]}"

    @property
    def items_type(self):
        """
        Return a list item type.

        Return the list type when a parameter type is
        of the form list[type], such as list[int]

        :return str:     The list item type
        """
        if not self._datatype:
            return None

        if m := re.search(r"(?:list|array)\[(\w+)\]", self._datatype):
            return TYPE_MAP[m.group(1)]

        return None

    def __repr__(self):
        """Return object representation."""
        return str(self.__dict__)

    @property
    def schema(self):
        """Return parameter schema."""
        _schema = {"type": self.datatype}
        if self.items_type:
            _schema["items"] = {"type": self.items_type}
        for x in ["enum", "min", "max", "default", "examples"]:
            if value := getattr(self, x):
                _schema[x] = value

        if self.in_request_body():
            _schema["description"] = self.description
            _schema["required"] = self.required

        return _schema

    def in_request_body(self):
        """
        Return if parameter is in request body.

        Returns `True` if parameter location is in the request body.

        :return bool:     True if in request body
        """
        if self.location in ["body", "form"]:
            return True

        return False

    def request_body(self):
        """
        Return request body.

        Returns the parameter request body schema.

        :return dict:      Request body
        """
        return self.schema

    def parameter(self):
        """
        Return parameter schema.

        Return the OpenAPI parameter schema.

        :return dict:     Parameter schema

        :return _type_: _description_
        """
        return {
            "name": self.name,
            "in": self.location,
            "description": self.description,
            "schema": self.schema,
            "required": self.required,
        }

    def dict(self):
        """
        Return the parameter dict.

        Returns the parameter dict for the correct parameter type.

        :return dict:     Parameter dict
        """
        if self.in_request_body():
            return self.request_body()

        return self.parameter()


class Response:
    """Response class."""

    def __init__(self, code=None, description=None, content=None, schema=None):
        """
        Create response instance.

        :param str code:          HTTP response code
        :param str description:   Description
        :param str content:       Content type
        :param dict schema:       Schema dictionary
        :return:
        """
        self.code = code
        self.description = description
        self.content = content
        self.schema = schema if schema else "default_response"

    def __repr__(self):
        """Return a printable representational string."""
        return str(self.__dict__)

    def dict(self):
        """Return dict of data."""
        content_types = [CONTENT_MAP.get(x) for x in self.content]
        schema = {"schema": {"$ref": f"#/components/schemas/{self.schema}"}}
        content = dict.fromkeys(content_types, schema)

        return {
            "description": self.description,
            "content": content,
        }


class Operation:
    """Operation Class."""

    def __init__(
        self,
        operation=None,
        summary=None,
        description=None,
        operation_id=None,
        parameters=None,
        tags=None,
        responses=None,
        callbacks=None,
        return_types=None,
        accepts=None,
        **_kwargs,
    ):
        """
        Create operation instance.

        :param str operation:           HTTP operation
        :param str summary:             Summary
        :param str description:         Description
        :param str operation_id:        Operation ID
        :param list parameters:         List of parameter dicts
        :param list tags:               List of tags
        :param list responses:          List of response dicts
        :param list callbacks:          List of callbacks
        :param list[str] return_types:  Return content type
        :param list[str] accepts:       Accept types
        :return:
        """
        self.operation = operation
        self.summary = summary
        self.description = description
        self.operation_id = operation_id
        self.return_types = return_types
        self.tags = tags or []
        self.callbacks = callbacks or {}
        self.parameters = [Parameter(**x) for x in parameters or []]
        self.responses = [Response(**x, content=return_types) for x in responses or []]
        self._accepts = accepts or ["*/*"]
        self.request_body_parameters = {x.name: x.dict() for x in self.parameters if x.in_request_body()}

    @property
    def accepts(self):
        """Return accepts type."""
        if self._accepts:
            return self._accepts
        if self.has_form():
            return ["form"]

        return ["*/*"]

    def binary_body(self):
        """Return binary body."""
        accepts = CONTENT_MAP.get(self.accepts)
        return {
            "content": {
                accepts: {
                    "schema": {
                        "type": "string",
                        "format": "binary",
                    }
                }
            }
        }

    def has_form(self):
        """
        Return whether operation has form data.

        Return `True` if the operation has form data parameters.

        :return bool:     True if has form data
        """
        return [x for x in self.parameters if x.location == "form"] != []

    def body(self):
        """Return request body."""
        accepts = [CONTENT_MAP.get(x) for x in self.accepts]
        required = [k for k, v in self.request_body_parameters.items() if v.get("required") is True]
        schema = {"schema": {"type": "object", "required": required, "properties": self.request_body_parameters}}
        content = dict.fromkeys(accepts, schema)

        return {
            "description": self.description,
            "content": content,
        }

    def request_body(self):
        """Return request body."""
        if self.accepts in BINARY_TYPES:
            return self.binary_body()

        return self.body()

    def dict(self):
        """Return dict of data."""
        operation = {
            "tags": self.tags,
            "summary": self.summary,
            "description": self.description,
            "operationId": self.operation_id,
            "parameters": [x.dict() for x in self.parameters if not x.in_request_body()],
            "responses": {x.code: x.dict() for x in self.responses},
        }
        if self.request_body_parameters:
            operation["requestBody"] = self.request_body()

        return {self.operation: operation}


class ResourceSchema:
    """Resource schema class."""

    def __init__(self, resource, parser=None):
        """
        Create resource instance.

        :param Resource resource:    Resource object
        :param Parser parser:        Doc string parser
        :return:
        """
        self.resource = resource
        self.routes = resource.__routes__
        self.name = resource.__class__.__name__.capitalize()
        self.tags = getattr(resource, "__tags__", [self.name.lower()])
        self.components = getattr(resource, "__components__", {})
        self.security = {verb.lower(): auth for verb, auth in getattr(resource, "__auth__", {}).items()}
        self.parser = parser() if parser else None
        self.paths = {}

    def methods(self):
        """
        Return the resource HTTP methods.

        Return the resources HTTP methods (get/put, etc)

        :return list:       List of methods
        """
        _methods = []
        for verb in verbs:
            method = getattr(self.resource, f"on_{verb}", None)
            if not method or not method.__doc__:
                continue
            _methods.append(method)

        return _methods

    def parse(self):
        """Parse the resource."""
        self.process_routes()

    def process_responses(self, operation):
        """
        Process operation responses.

        Process the responses of the operation.

        :param dict operation:     Operation dict
        :return dict:              Processed operation dict
        """
        for response in operation["responses"]:
            name = response["schema"]
            if name:
                schema = getattr(self.resource, name, {})
            else:
                schema = {response["code"]: response["description"]}

            if schema:
                self.components[name] = schema

    def process_parameters(self, operation):
        """
        Process operation parameters.

        Process the parameters of the operation.

        :param dict operation:     Operation dict
        :return dict:              Processed operation dict
        """
        for parameter in operation["parameters"]:
            enum = parameter.get("enum")
            parameter["enum"] = getattr(self.resource, enum) if enum else []

        return operation

    def process_routes(self):
        """
        Process the resource routes.

        Process the routes defined in the resource.

        :return:
        """
        for route in self.routes.keys():
            self.process_route(route)

    def process_route(self, route):
        """
        Process resource route.

        Process the operations of the resources route.

        :param str route:         The HTTP route
        :return:
        """
        self.paths[route] = {}
        paths = self.resource.__data__[route]

        # for each resource path add the path to openapi
        for v in paths.values():
            data = self.process_parameters(v)
            operation = Operation(tags=self.tags, **data)
            self.paths[route].update(operation.dict())
            self.process_responses(v)


class Info:
    """Info class."""

    def __init__(
        self,
        title=None,
        summary=None,
        description=None,
        terms=None,
        contact=None,
        license=None,
        version=None,
    ):
        """
        Create info class instance.

        :param str title:          API Title
        :param str summary:        Summary
        :param str description:    Description
        :param str terms:          Terms
        :param Contact contact:    Contact object
        :param License license:    License object
        :param str version:        API version
        :return:
        """
        self.title = title
        self.summary = summary
        self.description = description
        self.terms = terms
        self.contact = contact.dict()
        self.license = license.dict()
        self.version = version

    def __repr__(self):
        """Return a printable representational string."""
        return str(self.__dict__)

    def dict(self):
        """Return dict of data."""
        return {
            "title": self.title,
            "summary": self.summary,
            "description": self.description,
            "termsOfService": self.terms,
            "license": self.license,
            "version": self.version,
        }


class License:
    """License class."""

    def __init__(self, name=None, url=None):
        """
        Create license instance.

        :param str name:    License name
        :param are url:     License URL
        :return:
        """
        self.name = name or ""
        self.url = url or ""

    def dict(self):
        """Return dict of data."""
        return self.__dict__


class Contact:
    """API Contact information."""

    def __init__(self, name=None, url=None, email=None):
        """
        Create contact instance.

        :param str name:    Contact name
        :param str url:     Contact URL
        :param str email:   Contact email
        :return:
        """
        self.name = name
        self.url = url
        self.email = email

    def __repr__(self):
        """Return a printable representational string."""
        return str(self.__dict__)

    def dict(self):
        """Return dict of data."""
        return self.__dict__


class OpenApi:
    """
    OpenAPI schema class.

    Class representation of the OpenAPI specification.
    """

    def __init__(
        self,
        title="",
        description="",
        summary="",
        version="",
        terms="",
        license="",
        license_url="",
        contact_name="",
        contact_url="",
        contact_email="",
        auth=None,
        parser=None,
        servers=None,
    ):
        """
        Create an OpenAPI class.

        Creates and OpenAPI object that represents an OpenAPI
        specification document.

        :param str title:          API Title
        :param str description:    Description
        :param str summary:        Summary
        :param str version:        Version
        :param str terms:          Terms
        :param str license:        License
        :param str license_url:    URL to the License
        :param str contact_name:   Contact person
        :param str contact_url:    Software URL
        :param str contact_email:  Contact email
        :param Parser parser:      The docstring parser
        :param list servers:       List of servers
        :return:
        """
        self.openapi = "3.1.0"
        self.title = title or "<title>"
        self.description = description or "<description>"
        self.version = version or "0.0.0"
        self.summary = summary or ""
        self.terms = terms or ""
        self.auth = auth or []
        self.servers = servers or []

        contact = Contact(name=contact_name, email=contact_email, url=contact_url)
        license = License(name=license, url=license_url)
        self.info = Info(
            title=title,
            summary=summary,
            description=description,
            terms=terms,
            contact=contact,
            license=license,
            version=version,
        )
        self.paths = {}
        self.component_schemas = {
            "default_response": DEFAULT_RESPONSE,
            "binary": BINARY_RESPONSE,
        }
        self.parser = parser

    def process_resources(self, resources):
        """
        Generate OpenAPI data.

        Process and generate the OpenAPI data for the supplied resources.

        :param list[Resources] resources:     List of Resource objects
        :return:
        """
        for resource in resources:
            resource = ResourceSchema(resource)
            resource.parse()
            self.paths.update(resource.paths)
            self.component_schemas.update(resource.components)

    @property
    def security_schemes(self):
        """Return security schemas."""
        schema = {}
        for x in self.auth:
            schema.update(x.dict())

        return schema

    @property
    def security_names(self):
        """Return security list."""
        authenticators = [x for y in self.auth for x in y.authenticators]
        return [{x.name: []} for x in authenticators]

    @property
    def components(self):
        """Return components."""
        return {
            "securitySchemes": self.security_schemes,
            "schemas": self.component_schemas,
        }

    def schema(self):
        """Return OpenAPI Schema."""
        return {
            "openapi": self.openapi,
            "info": self.info.dict(),
            "paths": self.paths,
            "components": self.components,
            "security": self.security_names,
            "servers": self.servers,
        }
